/*
 * ADXF Library
 * 
 *   Copyright (C) 2016, Jonathan Sevy <jsevy@cs.drexel.edu>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */

package com.jsevy.adxftest;


import com.jsevy.adxf.BSpline;
import com.jsevy.adxf.DXFCanvas;
import com.jsevy.adxf.DXFDocument;
import com.jsevy.adxf.RealPoint;
import com.jsevy.adxf.SplineControlPoint;

import java.io.FileWriter;
import java.io.IOException;
import java.util.Vector;

import android.annotation.SuppressLint;
import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Matrix;
import android.graphics.Paint;
import android.graphics.Paint.Style;
import android.graphics.Path;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.Typeface;
import android.os.Environment;
import android.util.AttributeSet;
import android.util.SparseArray;
import android.view.MotionEvent;
import android.view.ScaleGestureDetector;
import android.view.View;

import com.jsevy.adxftest.R;


@SuppressLint("ClickableViewAccessibility")
public class DXFTestView extends View
						implements View.OnTouchListener
{
        
    public enum DisplayContent
    {
    	GRAPHICS,
    	TEXT,
    	TRANSFORMATIONS,
    	CLIPPING,
    	IMAGES,
    	BITMAP_MESH,
    	VERTICES_TRIANGLES,
    	VERTICES_FAN,
    	VERTICES_STRIP
    }
    
    private SparseArray<String> fileSuffixMap = new SparseArray<String>()
    {
        {
            put(R.id.testGraphicsButton, "graphics");
            put(R.id.testTextButton, "text");
            put(R.id.testTransformationsTextButton, "transformations_text");
            put(R.id.testTransformationsGraphicsButton, "transformations_graphics");
            put(R.id.testImagesButton, "images");
            put(R.id.testBitmapMeshButton, "mesh");
            put(R.id.testDrawVerticesTraingleButton, "vertices");
            put(R.id.testDrawVerticeFanButton, "vertices_fan");
            put(R.id.testDrawVerticesStripButton, "vertices_strip");
        }
    };
    
    
    // what to display and generate in the PDF file - just use button ID
    //private int displayContent = R.id.testGraphicsButton;
    private int displayContent = 0;
    
    private float displayScale = 1.0f;
    
    private float xCen = 0;
    private float yCen = 0;
    
    private float previousX = 0;
    private float previousY = 0;
    
    
    public void setDisplayContent(int displayContentButtonId)
    {
    	this.displayContent = displayContentButtonId;
    }
    
    
    
    public DXFTestView(Context context, AttributeSet attrs)
    {
        super(context, attrs);
        
        // set scale detector for pinch-stretch gesture handling
     	scaleDetector = new ScaleGestureDetector(this.getContext(), new ScaleListener());
     	
     	setOnTouchListener(this);
     	
     	// need this to get our onDraw method called...  :-\
     	this.setWillNotDraw(false);
     		
    }
    
    
    
    public void onDraw(Canvas canvas)
    {
        if (!(canvas instanceof DXFCanvas))
        {
        	// support for scrolling and resizing main view canvas to fit stuff on screen
            canvas.translate(xCen, yCen);
        	canvas.scale(displayScale,displayScale);
        }
    	drawDesign(canvas);
    }
    
    
    
    public void generatePDF()
    {
        DXFDocument dxfDocument = new DXFDocument("Example DXF output from Android DXFCanvas");
        DXFCanvas dxfCanvas = dxfDocument.getCanvas();
        
        drawDesign(dxfCanvas);
        
        try
        {
            // print dxf output to console for convenient inspection
            String stringOutput = dxfDocument.toDXFString();
            System.out.println(stringOutput);
            
            // now write to files - one with .dxf for display, and one with .txt for inspection
            // get appropriate suffix depending on what's generated
            String fileSuffix = fileSuffixMap.get(displayContent);
            if (fileSuffix == null)
            {
            	// hmm... not sure why we didn't find it - just leave off
            	fileSuffix = "";
            }
            
            String dxfPath = Environment.getExternalStoragePublicDirectory(Environment.DIRECTORY_DOWNLOADS).getAbsolutePath() + "/a_test_dxf_file_generation_" + fileSuffix + ".dxf";
            FileWriter dxfFileWriter = new FileWriter(dxfPath);
            dxfFileWriter.write(stringOutput);
            dxfFileWriter.flush();
            dxfFileWriter.close();
            
            // write a text file (*.txt) so can check out the generated pdf
            String textPath = Environment.getExternalStoragePublicDirectory(Environment.DIRECTORY_DOWNLOADS).getAbsolutePath() + "/a_test_dxf_file_generation_" + fileSuffix + ".txt";
            FileWriter textFileWriter = new FileWriter(textPath);
            textFileWriter.write(stringOutput);
            textFileWriter.flush();
            textFileWriter.close();
        }
        catch (IOException e)
        {
            System.out.println("Bah!!!!! " + e.toString());
        }
    }
    
    
    
    /**
     * Used to draw both on the screen and to the PDF document, to check if the PDF image matches the
     * one on the screen
     * @param canvas  Either a "normal" canvas for screen rendering or one associated with a PDF Page
     */
    private void drawDesign(Canvas canvas)
    {
        // draw some stuff
        Paint paint = new Paint();
        
        // fill the page with a pale yellow
        paint.setColor(Color.rgb(255, 245, 240));
        paint.setStyle(Style.FILL);
        canvas.drawPaint(paint);
        
        switch (displayContent)
        {
        
	        case R.id.testGraphicsButton:
	        {
	        	testGraphics(canvas);
	        	break;
	        }
	        
	        case R.id.testTextButton:
	        {
	        	testText(canvas);
	        	break;
	        }
	        
	        case R.id.testTransformationsGraphicsButton:
	        {
	        	testTransformationsGraphics(canvas);
	        	break;
	        }
	        
	        case R.id.testTransformationsTextButton:
	        {
	            testTransformationsText(canvas);
	        	break;
	        }
	             
	        default:
	        {
	        	// do nothing
	            //testGraphics(canvas);
	        	break;
	        }
	        
        
        };
        
    }
    
    
    
private ScaleGestureDetector scaleDetector;
	
	private class ScaleListener extends ScaleGestureDetector.SimpleOnScaleGestureListener 
	{
		
		@Override
		public boolean onScale(ScaleGestureDetector detector) 
		{
			float newDisplayScale = displayScale * detector.getScaleFactor();
			displayScale = newDisplayScale;
		    
		    invalidate();
		    
		    return true;
		}
	}
    
    
    
    @Override
	public boolean onTouch(View v, MotionEvent event) 
	{
		
		// Let the ScaleGestureDetector inspect all events.
	    scaleDetector.onTouchEvent(event);
	    
	    // see if we're moving a control point or the like
    	switch (event.getAction())
    	{
    		case MotionEvent.ACTION_DOWN:
    		{
    			previousX = event.getX();
    			previousY = event.getY();
    			
    			break;
    		}
    		
    		case MotionEvent.ACTION_MOVE:
    		{
    			float deltaX = event.getX() - previousX;
    			float deltaY = event.getY() - previousY;
    			
    			xCen += deltaX;
    			yCen += deltaY;
    			
    			previousX = event.getX();
    			previousY = event.getY();
    			
    			invalidate();
    			
    			break;
    		}
    		
    		case MotionEvent.ACTION_UP:
    		{
    			
    			break;
    		}
    		
    		default:
    		{
    			break;
    		}
    	}
		
		return true;
	    
    	
	}
    
    
    
   
    
    
    
    @SuppressWarnings("deprecation")
    private void testText(Canvas canvas)
    {
        // do some stuff
        Paint paint = new Paint();
        
        paint.setTypeface(Typeface.SERIF);
        paint.setColor(Color.RED);
        paint.setTextSize(24);
        paint.setStyle(Style.STROKE);
        
        canvas.drawLine(0, 0, 150, 250, paint);
        canvas.drawText("Some 24-point serif red text at position 150, 250 with some accented characters: áéíóúüèìç", 150, 250, paint);
        
        paint.setTextSize(48);
        paint.setColor(Color.GREEN);
        paint.setTypeface(Typeface.SANS_SERIF);
        canvas.drawLine(0, 0, 180, 350, paint);
        canvas.drawText("Some 48-point sans-serif green text at position 180, 350", 180, 350, paint);
        
        paint.setTextSize(38);
        paint.setColor(Color.BLUE);
        paint.setTypeface(Typeface.MONOSPACE);
        canvas.drawLine(0, 0, 480, 400, paint);
        canvas.drawText("Some 38-point monospace blue text at position 480, 400", 480, 400, paint);
        
        // try drawing characters at indicated positions
        paint.setTypeface(Typeface.create(Typeface.SERIF, Typeface.ITALIC));
        float[] pos = { 400, 500, 440, 510, 480, 525, 520, 545, 560, 570, 600, 600, 640, 635, 680, 675, 720, 720, 760, 770, 800, 825};
        String text = "drawPosText";
        canvas.drawPosText(text, pos, paint);
        
        // draw characters along path
        paint.setTypeface(Typeface.create(Typeface.SERIF, Typeface.BOLD));
        Path path = new Path();
        path.addArc(new RectF(100, 1000, 700, 1400), 180, -180);
        canvas.drawArc(new RectF(100, 1000, 700, 1400), 180, -180, false, paint);
        path.arcTo(new RectF(700, 1000, 1400, 1400), 180, 180);
        canvas.drawArc(new RectF(700, 1000, 1400, 1400), 180, 180, false, paint);
        path.arcTo(new RectF(1400, 1100, 1600, 1300), 180, -180);
        canvas.drawArc(new RectF(1400, 1100, 1600, 1300), 180, -180, false, paint);
        text = "Example of bold serif drawTextOnPath with offset and composite path";
        canvas.drawTextOnPath(text, path, 100, 40, paint);
        
    }
    
    
    
    private void testGraphics(Canvas canvas)
    {
        // do some stuff
        Paint paint = new Paint();
        
        // draw some stuff in blue
        paint.setColor(Color.BLUE);
        paint.setStrokeWidth(5);
        paint.setStyle(Style.STROKE);
        
        canvas.drawLine(0, 0, 800, 800, paint);
        canvas.drawCircle(800, 800, 170, paint);
        canvas.drawCircle(800, 800, 10, paint);
        canvas.drawArc(new RectF(750, 750, 850, 850), 180, -90, false, paint);
        
        // fill a rectangle
        paint.setColor(Color.RED);
        paint.setStrokeWidth(3);
        paint.setStyle(Style.FILL);
        canvas.drawLine(0, 0, 1000, 500, paint);
        canvas.drawRect(1000, 500, 1150, 650, paint);
        
        // draw a rounded rectangle
        paint.setStyle(Style.STROKE);
        canvas.drawLine(0, 0, 20, 200, paint);
        canvas.drawRoundRect(new RectF(20, 200, 150, 300), 20, 10, paint);
        
        // draw some ovals
        paint.setColor(Color.MAGENTA);
        paint.setStrokeWidth(15);
        canvas.drawLine(0, 0, 300, 1000, paint);
        canvas.drawOval(new RectF(200, 800, 400, 1200), paint);
        canvas.drawRect(200, 800, 400, 1200, paint);
        
        // draw oval arc
        paint.setStrokeWidth(1);
        canvas.drawLine(0, 0, 300, 2000, paint);
        //canvas.drawOval(new RectF(100, 1900, 500, 2100), paint);
        canvas.drawArc(new RectF(100, 1900, 500, 2100), 60, 150, true, paint);
        canvas.drawRect(100, 1900, 500, 2100, paint);
        
        // now draw a spline; static BSpline method handles both regular and DXF canvas - since no native
        // splines in standard canvas, spline is drawn as a set of connected line segments, with the tolerance
        // parameter indicating that maximum length of any line segment to control smoothness of the representation
        {
            int degree = 3;
            float[] controlPoints = {500, 1600, 500, 1300, 700, 1800, 800, 1200, 1000, 2000, 500, 1200};
            int[] multiplicities = {3, 1, 1, 2, 1, 3};
            
            BSpline.drawSpline(canvas, degree, controlPoints, multiplicities, true, 1, paint);
        }
        
        // draw a series of splines using the same control points but with different degrees
        {
            
            float[] controlPoints = {1000, 1600, 1000, 1300, 1200, 1800, 1300, 1200, 1500, 2000, 1000, 1200, 1400, 1200};
            int[] multiplicities = {1, 1, 1, 1, 1, 1, 1};
            
            for (int degree = 2; degree < 6; degree++)
            {
                BSpline.drawSpline(canvas, degree, controlPoints, multiplicities, true, 1, paint);
            }
        }
        
    }
    
    
    
    private void testTransformationsGraphics(Canvas canvas)
    {
        // do some stuff
        Paint paint = new Paint();
        paint.setStyle(Style.STROKE);
        //paint.setStrokeWidth(3);
        
        // save state to restore
        canvas.save();
        
        // Rotation: draw a series of rotated elliptical arcs
        paint.setColor(Color.GREEN);
        
        for (int i = 0; i < 40; i += 5)
        {
            canvas.rotate(-5); // counterclockwise...
            canvas.drawLine(0, 0, 250, 1550, paint);
            //canvas.drawOval(new RectF(100, 1500, 400, 1600), paint);
            canvas.drawArc(new RectF(100, 1500, 400, 1600), 60, 150, true, paint);
        }
        
        canvas.restore();
        canvas.save();
        
        // Skew: draw a series of increasingly skewed circular arcs and rectangles
        paint.setColor(Color.BLUE);
        for (int i = 0; i < 8; i++)
        {
        	canvas.drawLine(0, 0, 200, 200, paint);
            //canvas.drawOval(new RectF(100, 1500, 400, 1600), paint);
            canvas.drawArc(new RectF(100, 100, 300, 300), 45, 135, true, paint);
            canvas.skew(0.2f, 0.0f);
        }
        
        canvas.restore();
        canvas.save();
        
        paint.setColor(Color.CYAN);
        for (int i = 0; i < 8; i++)
        {
        	canvas.drawLine(0, 0, 100, 100, paint);
            canvas.drawRect(new RectF(100, 100, 300, 300), paint);
            canvas.skew(0.1f, 0.2f);
        }
        
        canvas.restore();
        canvas.save();
        
        paint.setColor(Color.RED);
        // Scaling: do series of scaled circular arcs
        for (int i = 0; i < 8; i++)
        {
        	canvas.drawLine(0, 0, 500, 200, paint);
            //canvas.drawCircle(500, 200, 50, paint);
            canvas.drawArc(new RectF(450, 150, 550, 250), -45, 135, true, paint);
            
            canvas.drawLine(0, 0, 200, 500, paint);
            //canvas.drawCircle(500, 200, 50, paint);
            canvas.drawArc(new RectF(150, 450, 250, 550), 80, -60, true, paint);
            canvas.scale(1.1f, -1.2f);
        }
        
        canvas.restore();
        canvas.save();
        
        // Translation: do series of translated circular arcs
        paint.setColor(Color.MAGENTA);
        for (int i = 0; i < 8; i++)
        {
        	canvas.drawLine(0, 0, 700, -350, paint);
            //canvas.drawCircle(500, 200, 50, paint);
            canvas.drawArc(new RectF(600, -450, 800, -250), -135, 315, true, paint);
            canvas.translate(200, 100);
        }
        
        canvas.restore();
        canvas.save();
        
        // combine skew and rotate and translate
        paint.setColor(Color.BLACK);
        canvas.translate(100, 300);
        canvas.rotate(-40); // counterclockwise...
        canvas.skew(-0.4f, -0.3f);
        canvas.drawLine(0, 0, 80, 950, paint);
        RectF rect = new RectF(80, 880, 700, 1020);
        canvas.drawRect(rect, paint);
        canvas.drawArc(rect, -45, 270, true, paint);
        
    }
    
    
    
    private void testTransformationsText(Canvas canvas)
    {
        // do some stuff
        Paint paint = new Paint();
        
        // draw a rotated rectangle and some text
        paint.setStyle(Style.STROKE);
        paint.setStrokeWidth(3);
        paint.setTypeface(Typeface.SERIF);
        paint.setTextSize(40);
        paint.setColor(Color.GREEN);
        
        // save state to restore
        canvas.save();
        
        for (int i = 0; i < 40; i += 5)
        {
            canvas.rotate(-5); // counterclockwise...
            canvas.drawLine(0, 0, 250, 1550, paint);
            //canvas.drawOval(new RectF(100, 1500, 400, 1600), paint);
            canvas.drawArc(new RectF(100, 1500, 400, 1600), 60, 150, true, paint);
        }
        
        canvas.restore();
        
        paint.setColor(Color.BLUE);
        // canvas.scale(2, 2);
        canvas.skew(0.3f, 0);
        canvas.drawLine(0, 0, 75, 380, paint);
        canvas.drawRect(75, 380, 700, 520, paint);
        canvas.drawText("Some 40-point blue text skewed at position 80, 450", 80, 450, paint);
        
        // canvas.setMatrix(origMatrix);
        canvas.skew(-0.3f, 0);
        
        paint.setColor(Color.BLUE);
        canvas.skew(0, 0.8f);
        canvas.drawLine(0, 0, 80, 950, paint);
        canvas.drawRect(80, 880, 700, 1020, paint);
        canvas.drawText("Some 40-point blue text skewed at position 80, 950", 80, 950, paint);
        
        canvas.skew(0, -0.8f);
        
        // both skew and rotate and translate
        paint.setColor(Color.RED);
        canvas.translate(100, 300);
        canvas.rotate(-40); // counterclockwise...
        canvas.skew(-0.4f, -0.3f);
        canvas.drawLine(0, 0, 80, 950, paint);
        canvas.drawRect(80, 880, 700, 1020, paint);
        canvas.drawText("Some 40-point red text translated (100,300), rotated -40, and skewed, at position 80, 950", 80, 950, paint);
        
    }
    
    
    
    /**
     * Helper class for some of the grid-related tests
     * 
     * @author jsevy
     *
     */
    public class Vector3D
    {
        float x;
        float y;
        float z;
        
        
        
        public Vector3D(float x, float y, float z)
        {
            this.x = x;
            this.y = y;
            this.z = z;
        }
        
        
        
        public Vector3D()
        {
            this.x = 0;
            this.y = 0;
            this.z = 0;
        }
        
        
        
        public float dotProduct(Vector3D other)
        {
            return this.x * other.x + this.y * other.y + this.z * other.z;
        }
        
        
        
        public float magnitude()
        {
            return (float) Math.sqrt(this.dotProduct(this));
        }
        
        
        
        public Vector3D scalarProduct(float scale)
        {
            Vector3D result = new Vector3D();
            
            result.x = this.x * scale;
            result.y = this.y * scale;
            result.z = this.z * scale;
            
            return result;
        }
        
        
        
        public Vector3D crossProduct(Vector3D other)
        {
            Vector3D result = new Vector3D();
            
            result.x = this.y * other.z - other.y * this.z;
            result.y = this.x * other.z - other.x * this.z;
            result.z = this.x * other.y - other.x * this.y;
            
            return result;
        }
        
        
        
        public Vector3D sum(Vector3D other)
        {
            Vector3D result = new Vector3D();
            
            result.x = this.x + other.x;
            result.y = this.y + other.y;
            result.z = this.z + other.z;
            
            return result;
        }
    }
}
